"""
Module for defining and managing the event lifecycle in the LLM Compressor.

This module provides an abstract base class for defining event lifecycles
and methods for retrieving events based on their type,
managing step and batch counts, and triggering events.
"""

from abc import ABC, abstractmethod
from typing import List

from loguru import logger

from llmcompressor.core.events.event import Event, EventType

__all__ = [
    "EventLifecycle",
]


class EventLifecycle(ABC, Event):
    """
    A lifecycle for events to be used in a LLMCompressor session.
    Provides base utilities and defines the contract that
    all inheritors must follow.

    The order in which the events are called is determined by
    the inheritors of this class.

    The expected lifecycle is as follows with optional gradient accumulation:
        for gradient_batches in training loop:
            for batch in gradient_batches or [gradient_batches]:
                BATCH_START
                LOSS_CALCULATED

                if not last batch:
                    BATCH_END
                else:
                    OPTIM_PRE_STEP
                    OPTIM_POST_STEP
                    BATCH_END
    For older flows where the optimizer is wrapped and invocations_per_step > 1:
        for gradient_batches in training loop:
            for invocation in range(invocations_per_step):
                for batch in gradient_batches or [gradient_batches]:
                    BATCH_START
                    LOSS_CALCULATED

                    if not last batch or not last invocation:
                        BATCH_END
                    else:
                        OPTIM_PRE_STEP
                        OPTIM_POST_STEP
                        BATCH_END

    :param type_first: The first event type to be called
    :type type_first: EventType
    :param start: The start event to base the lifecycle off of
    :type start: Event
    """

    def __init__(self, type_first: EventType, start: Event):
        logger.debug(
            "Initializing EventLifecycle with type_first={} and start={}",
            type_first,
            start,
        )
        self.type_first = type_first
        self.last_type = None

        self.steps_per_epoch = start.steps_per_epoch
        self.batches_per_step = start.batches_per_step
        self.invocations_per_step = start.invocations_per_step
        self.global_step = start.global_step
        self.global_batch = start.global_batch

    def events_from_type(self, type_: EventType) -> List[Event]:
        """
        Get the list of events for a given type.

        :param type_: The event type to get the events for
        :type type_: EventType
        :return: The list of events for the given type
        :rtype: List[Event]
        :raises ValueError: If the event type is invalid
        """
        logger.debug("Fetching events from type: {}", type_)
        if type_ == EventType.BATCH_START:
            return self.batch_start_events()
        if type_ == EventType.LOSS_CALCULATED:
            return self.loss_calculated_events()
        if type_ == EventType.OPTIM_PRE_STEP:
            return self.optim_pre_step_events()
        if type_ == EventType.OPTIM_POST_STEP:
            return self.optim_post_step_events()
        if type_ == EventType.BATCH_END:
            return self.batch_end_events()
        logger.error("Invalid event type: {}", type_)
        raise ValueError(f"Invalid event type {type_}")

    def check_batches_per_step_count(self, increment: bool) -> bool:
        """
        Check if the batch count is at the step or step invocation count.
        If batches_per_step is None or < 2, always returns True.

        If invocations_per_step is > 1,
        then returns True for batches matching the invocation.
        Check check_invocations_per_step_count for the invocation count.

        :param increment: Whether to increment the batch count
        :type increment: bool
        :return: True if the batch count is at the step count, False otherwise
        :rtype: bool
        """
        compare_batch = self.global_batch + 1
        at_step = (
            self.batches_per_step is None
            or self.batches_per_step < 2
            or (compare_batch % self.batches_per_step == 0)
        )
        if increment:
            self.global_batch = compare_batch

        return at_step

    def check_invocations_per_step_count(self, increment: bool) -> bool:
        """
        Check if the invocation count is at the step count.
        If invocations_per_step is None or < 2, always returns True.

        :param increment: Whether to increment the step count
        :type increment: bool
        :return: True if the invocation count is at the step count, False otherwise
        :rtype: bool
        """
        compare_step = self.global_step + 1
        at_step = (
            self.invocations_per_step is None
            or self.invocations_per_step < 2
            or (compare_step % self.invocations_per_step == 0)
        )

        if increment:
            self.global_step = compare_step

        return at_step

    @abstractmethod
    def batch_start_events(self) -> List[Event]:
        """Return the list of events to be called for the batch start."""
        raise NotImplementedError()

    @abstractmethod
    def loss_calculated_events(self) -> List[Event]:
        """Return the list of events to be called for the loss calculated."""
        raise NotImplementedError()

    @abstractmethod
    def optim_pre_step_events(self) -> List[Event]:
        """Return the list of events to be called for the optim pre step."""
        raise NotImplementedError()

    @abstractmethod
    def optim_post_step_events(self) -> List[Event]:
        """Return the list of events to be called for the optim post step."""
        raise NotImplementedError()

    @abstractmethod
    def batch_end_events(self) -> List[Event]:
        """Return the list of events to be called for the batch end."""
        raise NotImplementedError()



# flake8: noqa
# coding=utf-8
# Copyright 2022 The OpenAI Authors and The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# vllm-project: no copyright
"""PyTorch Whisper model."""

import torch
from torch import nn

from transformers import WhisperConfig
from transformers.models.whisper.modeling_whisper import (
    WhisperEncoder,
    WhisperDecoder,
    WhisperModel,
    WhisperForConditionalGeneration,
    WhisperForAudioClassification,
)
from transformers.modeling_outputs import BaseModelOutput


class WhisperEncoder(WhisperEncoder):
    def forward(
        self,
        input_features,
        attention_mask=None,
        head_mask=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):

        expected_seq_length = self.config.max_source_positions * self.conv1.stride[0] * self.conv2.stride[0]
        # TRACING: assume preprocessing is correct
        # if input_features.shape[-1] != expected_seq_length:
        if False:
            raise ValueError(
                f"Whisper expects the mel input features to be of length {expected_seq_length}, but found {input_features.shape[-1]}. Make sure to pad the input mel features to {expected_seq_length}."
            )

        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict
        inputs_embeds = nn.functional.gelu(self.conv1(input_features))
        inputs_embeds = nn.functional.gelu(self.conv2(inputs_embeds))

        inputs_embeds = inputs_embeds.permute(0, 2, 1)
        embed_pos = self.embed_positions.weight

        hidden_states = inputs_embeds + embed_pos
        hidden_states = nn.functional.dropout(hidden_states, p=self.dropout, training=self.training)

        encoder_states = () if output_hidden_states else None
        all_attentions = () if output_attentions else None

        # check if head_mask has a correct number of layers specified if desired
        if head_mask is not None:
            assert head_mask.size()[0] == (
                len(self.layers)
            ), f"The head_mask should be specified for {len(self.layers)} layers, but it is for {head_mask.size()[0]}."

        for idx, encoder_layer in enumerate(self.layers):
            if output_hidden_states:
                encoder_states = encoder_states + (hidden_states,)
            # add LayerDrop (see https://arxiv.org/abs/1909.11556 for description)
            to_drop = False
            if self.training:
                dropout_probability = torch.rand([])
                if dropout_probability < self.layerdrop:  # skip the layer
                    to_drop = True

            if to_drop:
                layer_outputs = (None, None)
            else:
                if self.gradient_checkpointing and self.training:
                    layer_outputs = self._gradient_checkpointing_func(
                        encoder_layer.__call__,
                        hidden_states,
                        None,
                        (head_mask[idx] if head_mask is not None else None),
                        output_attentions,
                    )
                else:
                    layer_outputs = encoder_layer(
                        hidden_states,
                        None,
                        layer_head_mask=(head_mask[idx] if head_mask is not None else None),
                        output_attentions=output_attentions,
                    )

                hidden_states = layer_outputs[0]

            if output_attentions:
                all_attentions = all_attentions + (layer_outputs[1],)

        hidden_states = self.layer_norm(hidden_states)
        if output_hidden_states:
            encoder_states = encoder_states + (hidden_states,)

        if not return_dict:
            return tuple(v for v in [hidden_states, encoder_states, all_attentions] if v is not None)
        return BaseModelOutput(
            last_hidden_state=hidden_states, hidden_states=encoder_states, attentions=all_attentions
        )


class WhisperModel(WhisperModel):
    def __init__(self, config: WhisperConfig):
        super().__init__(config)

        self.encoder = WhisperEncoder(config)
        self.decoder = WhisperDecoder(config)
        # Initialize weights and apply final processing
        self.post_init()

class WhisperForConditionalGeneration(WhisperForConditionalGeneration):
    def __init__(self, config: WhisperConfig):
        super().__init__(config)
        self.model = WhisperModel(config)
        self.proj_out = nn.Linear(config.d_model, config.vocab_size, bias=False)
        self.max_target_positions = config.max_target_positions

        # Initialize weights and apply final processing
        self.post_init()


class WhisperForAudioClassification(WhisperForAudioClassification):
    def __init__(self, config):
        super().__init__(config)

        self.encoder = WhisperEncoder(config)
        num_layers = config.num_hidden_layers + 1  # transformer layers + input embeddings
        if config.use_weighted_layer_sum:
            self.layer_weights = nn.Parameter(torch.ones(num_layers) / num_layers)
        self.projector = nn.Linear(config.hidden_size, config.classifier_proj_size)
        self.classifier = nn.Linear(config.classifier_proj_size, config.num_labels)

        # Initialize weights and apply final processing
        self.post_init()